"""
This module contains various classes, functions and models which are shared across the app, particularly by invocations.

Lifting these classes, functions and models into this shared module helps to reduce circular imports.
"""
