import{r as o,O as i,t8 as t,t9 as n,aB as r,ta as v,tb as a,tc as c}from"./index-D0KA5Hsh.js";function l({children:s}){const{i18n:p}=i(),e=p.dir(),x=o.useMemo(()=>t({...n,direction:e,shadows:{...n.shadows,selected:"inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-500), inset 0px 0px 0px 4px var(--invoke-colors-invokeBlue-800)",hoverSelected:"inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-400), inset 0px 0px 0px 4px var(--invoke-colors-invokeBlue-800)",hoverUnselected:"inset 0px 0px 0px 2px var(--invoke-colors-invokeBlue-300), inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-800)",selectedForCompare:"inset 0px 0px 0px 3px var(--invoke-colors-invokeGreen-300), inset 0px 0px 0px 4px var(--invoke-colors-invokeGreen-800)",hoverSelectedForCompare:"inset 0px 0px 0px 3px var(--invoke-colors-invokeGreen-200), inset 0px 0px 0px 4px var(--invoke-colors-invokeGreen-800)"}}),[e]);return o.useEffect(()=>{document.body.dir=e},[e]),r.jsx(v,{theme:x,toastOptions:a,children:r.jsx(c,{children:s})})}const d=o.memo(l);export{d as default};
